#include "IntCache.H"


IntCache::IntCache(unsigned int size):
  size_(size), length_(0), buffer_(new unsigned int[size]),
  lastValueInserted_(0), lastDiff_(0)
{
}


int
IntCache::lookup(unsigned int& value, unsigned int& index, unsigned int mask,
		 unsigned int& sameDiff)
{
  unsigned int minDiff = 0xffffffff;
  unsigned int minDiffIndex;
  for (unsigned int i = 0; i < length_; i++)
    if (value == buffer_[i])
    {
      index = i;
      if (i)
      {
	unsigned int target = (i >> 1);
	do
	{
	  buffer_[i] = buffer_[i - 1];
	  i--;
	} while (i > target);
	buffer_[target] = value;
      }
      return 1;
    }
    else
    {
      unsigned int diff = ((value > buffer_[i]) ? value - buffer_[i] :
			   buffer_[i] - value);
      if (diff < minDiff)
      {
	minDiff = diff;
	minDiffIndex = i;
      }
    }

  unsigned int insertionPoint;
  if (2 >= length_)
    insertionPoint = length_;
  else
    insertionPoint = 2;
  unsigned int start;
  if (length_ >= size_)
    start = size_ - 1;
  else
  {
    start = length_;
    length_++;
  }
  for (unsigned int k = start; k > insertionPoint; k--)
    buffer_[k] = buffer_[k - 1];
  buffer_[insertionPoint] = value;
  unsigned int diff = value - lastValueInserted_;

  lastValueInserted_ = (value & mask);
  value = (diff & mask);
  sameDiff = (value == lastDiff_);
  lastDiff_ = value;

  return 0;
}


unsigned int
IntCache::get(unsigned int index)
{
  unsigned int result = buffer_[index];
  if (index != 0)
  {
    unsigned int i = index;
    unsigned int target = (i >> 1);
    do
    {
      buffer_[i] = buffer_[i - 1];
      i--;
    } while (i > target);
    buffer_[target] = result;
  }

  return result;
}


void
IntCache::insert(unsigned int& value, unsigned int mask)
{
  unsigned int insertionPoint;
  if (2 >= length_)
    insertionPoint = length_;
  else
    insertionPoint = 2;
  unsigned int start;
  if (length_ >= size_)
    start = size_ - 1;
  else
  {
    start = length_;
    length_++;
  }
  for (unsigned int k = start; k > insertionPoint; k--)
    buffer_[k] = buffer_[k - 1];
  lastDiff_ = value;
  lastValueInserted_ += value;
  lastValueInserted_ &= mask;
  buffer_[insertionPoint] = lastValueInserted_;
  value = lastValueInserted_;
}


unsigned int
IntCache::getBlockSize(unsigned int numBits) const
{
  unsigned int lastChangeIndex = 0;
  unsigned int lastBitIsOne = (lastDiff_ & 0x1);
  unsigned int mask = 0x2;
  for (unsigned int i = 1; i < numBits; i++)
  {
    unsigned int nextBitIsOne = (lastDiff_ & mask);
    if ((nextBitIsOne && !lastBitIsOne) || (lastBitIsOne && !lastBitIsOne))
      lastChangeIndex = i;
    lastBitIsOne = nextBitIsOne;
    mask <<= 1;
  }
  unsigned int result = lastChangeIndex + 1;
  if (result < 2)
    result = 2;
  return result;
}
